/*----------------------------------------------
* mrchexit.c - Source file for a C DLL
*----------------------------------------------*/
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <cmqc.h>
#include <cmqxc.h>

#ifdef _WIN32
#include <winsock2.h>
/*#include <windows.h>*/
#endif

#include "mqchexit.h"

#ifdef _AIX
#include <syslog.h>
#endif

/* Names of environment variables used to locate key files and log files */
#define KEYFILE_PATH_VAR "MQSCY_KEYFILEPATH"
#define LOGFILE_PATH_VAR "MQSCY_LOGFILEPATH"

/* Trap messages. */
#define NORMAL_TRAP 1100
#define CRITICAL_TRAP 1200
#define MQS_HAM_BUFLEN 1024

static FILE *logFilePtr = 0;
static char logFileName[256] = "d:\\mqm\\exits\\mqsec.log";
static char logFilePath[256];
static int doDebugging = 0;
static char keyFilePath[256];
unsigned char privKeyFileName[256];
unsigned char pubKeyFileName[256];
static char hostName[256];
unsigned int hostNameLen;
MQ_QMGR_DATA qmgrdata;
MQ_CHANNEL_DATA channeldata;
R_RSA_PUBLIC_KEY cssdPubk;
R_RSA_PUBLIC_KEY partnerpubk;
MQCHAR partnername[MQ_CHANNEL_NAME_LENGTH + 1];
MQ_HAM origMqhamMsg;
MQ_HAM mqhamMsg;
unsigned char tmpBuf[1024];

/* Prototypes */
void MQENTRY ChannelExit (PMQVOID channelExitParms,
                          PMQVOID channelDef,
                          PMQLONG dataLength,
                          PMQLONG agBufLength,
                          PMQVOID agBuf,
                          PMQLONG exitBufLength,
                          PMQPTR exitBufAddr);

void MQENTRY SecurityChannelExit (PMQVOID channelExitParms,
                                  PMQVOID channelDef,
                                  PMQLONG dataLength,
                                  PMQLONG agBufLength,
                                  PMQVOID agBuf,
                                  PMQLONG exitBufLength, PMQPTR exitBufAddr);

void MQENTRY SendMessageExit (PMQVOID channelExitParms,
                              PMQVOID channelDef,
                              PMQLONG dataLength,
                              PMQLONG agBufLength,
                              PMQVOID agBuf,
                              PMQLONG exitBufLength, PMQPTR exitBufAddr);

void MQENTRY ReceiveMessageExit (PMQVOID channelExitParms,
                                 PMQVOID channelDef,
                                 PMQLONG dataLength,
                                 PMQLONG agBufLength,
                                 PMQVOID agBuf,
                                 PMQLONG exitBufLength, PMQPTR exitBufAddr);

void MQStart (void);

int ReadKeysFromFiles (void);

int ReadPrivateKey (void);

int ReadPublicKey (void);

unsigned char *AllocExitBuffer (int len);

void FreeExitBuffer (PMQPTR exitBufAddr, PMQLONG exitBufLength);

void MakeCString (char *bf, char *zt, int len);

void GetInfo (PMQCD pChDef);

int GetHostName (unsigned char *buf, unsigned int *bufLen);

void BuildLogFileName (void);

void WriteLogFile (int msgid, unsigned char msgtype, unsigned char *msgtxt, ...);

void PrintBuff (unsigned char *out, unsigned int len);

int BuildHAM (char *buf, unsigned int *bufLen, MQ_HAM * mqham);

int ReceiveHAM (MQ_HAM * mqham, char *buf, unsigned int buflen);

int ReadAnswerHAM (MQ_HAM * mqham, char *buf, unsigned int buflen);

int BuildBufferToSendHAM (unsigned char *outbuf, unsigned int *outbuflen, MQ_HAM * mqham);

int EncryptHAM (MQ_HAM * mqham, R_RSA_PUBLIC_KEY * pubk, R_RSA_PRIVATE_KEY * privk);

int InitHAM (MQ_HAM * mqham);

int EncryptBuffer (unsigned char *outbuf,
                   unsigned int *outbuflen,
                   unsigned char *inbuf,
                   unsigned int inbuflen,
                   unsigned int compressflag,
                   R_RSA_PUBLIC_KEY * recpubk,
                   R_RSA_PRIVATE_KEY * sdrprivk,
                   MQ_DIGITAL_SIGNATURE * mqdigsig);

int DecryptAndVerifyBuffer (unsigned char *outbuf,
                            unsigned int *outbuflen,
                            unsigned char *inbuf,
                            unsigned int inbuflen,
                            R_RSA_PUBLIC_KEY * sdrpubk,
                            R_RSA_PRIVATE_KEY * recprivk,
                            MQ_DIGITAL_SIGNATURE * mqdigsig);

int FillRandomBuffer (unsigned char *outbuf, int len);

int FillRandomBuffer16 (unsigned char *outbuf);

int ReadHAM (MQ_HAM * mqham, unsigned char *rcvrbuf, unsigned int rcvrbuflen);

int VerifyHAM (MQ_HAM * mqham, R_RSA_PUBLIC_KEY * pubk, R_RSA_PRIVATE_KEY * privk);
int VerifyAnswerHAM (MQ_HAM * mqham, MQ_HAM * origmqham, R_RSA_PUBLIC_KEY * sdrpubk,
                     R_RSA_PRIVATE_KEY * rcvrprivk);

unsigned char *BuildBufferToSendMSG (unsigned int *outbuflen,
                                     int *status,
                                     unsigned char *inbuf,
                                     unsigned int inbuflen);

unsigned char *ReceiveMSGBuffer (unsigned int *outbuflen,
                                 unsigned char *inbuf,
                                 int *status, unsigned int inbuflen);

/* This is the actual ChannelExit routine which is exported and
* will be invoked from MQSeries.
* It acts as a dispatcher to call the approriate exit routines:
* SecurityChannelExit()
* SendMessageExit()
* ReceiveMessageExit()
*
*/
void /*_Export*/ MQENTRY ChannelExit (PMQVOID channelExitParms,
                                  PMQVOID channelDef,
                                  PMQLONG dataLength,
                                  PMQLONG agBufLength,
                                  PMQVOID agBuf,
                                  PMQLONG exitBufLength, PMQPTR exitBufAddr)
{
  PMQCXP pChlExParms = (PMQCXP) channelExitParms;
  PMQCD pChDef = (PMQCD) channelDef;
  static int firstTime = 1;
  if (firstTime) {
    GetInfo (pChDef);
    BuildLogFileName ();
    firstTime = 0;
  }
  switch (pChlExParms->ExitId) {
  case MQXT_CHANNEL_SEC_EXIT:
    WriteLogFile (400, 'I', ">>>>> Security exit started\n");
    SecurityChannelExit (channelExitParms,
                         channelDef,
                         dataLength,
                         agBufLength, agBuf, exitBufLength, exitBufAddr);
    break;
  case MQXT_CHANNEL_SEND_EXIT:
    WriteLogFile (410, 'I', ">>>>> Send message exit started\n");
    SendMessageExit (channelExitParms,
                     channelDef,
                     dataLength,
                     agBufLength, agBuf, exitBufLength, exitBufAddr);
    break;
  case MQXT_CHANNEL_RCV_EXIT:
    WriteLogFile (420, 'I', ">>>>> Receive message exit started\n");
    ReceiveMessageExit (channelExitParms,
                        channelDef,
                        dataLength,
                        agBufLength, agBuf, exitBufLength, exitBufAddr);
    break;
  case MQXT_CHANNEL_MSG_EXIT:
    WriteLogFile (1, 'E', "Channel message exit will not be handled here\n");
  default:;
  }                                /* endswitch ExitType */
}

/*
* This is the function for the channel security exit
*/
void MQENTRY SecurityChannelExit (PMQVOID channelExitParms,
                                  PMQVOID channelDef,
                                  PMQLONG dataLength,
                                  PMQLONG agBufLength,
                                  PMQVOID agBuf,
                                  PMQLONG exitBufLength, PMQPTR exitBufAddr)
{
  int i = 0;
  unsigned long ExitBufferLen = 0;
  char *pExitBuffer = (char *) agBuf;
  char *pagBuf = (char *) agBuf;
  PMQCXP pChlExParms = (PMQCXP) channelExitParms;
  PMQCD pChDef = (PMQCD) channelDef;
  time_t datetime;
  int status;
  char *buf;
  unsigned int bufLen;
  MQStart ();
  FreeExitBuffer (exitBufAddr, exitBufLength);
  switch (pChlExParms->ExitReason) {
  case MQXR_INIT:
    WriteLogFile (600, 'D', "ExitReason = MQXR_INIT\n");
    status = ReadKeysFromFiles ();
    if (status != 0) {
      WriteLogFile (12, 'E',
                    "Reading keys from file unsuccessful. Error code: 0x%04lX\n",
                    status);
      pChlExParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
      pChlExParms->Feedback = 0L;
      pChlExParms->ExitResponse2 = 0L;
      break;
    }
    pChlExParms->ExitResponse2 = 0L;
    pChlExParms->Feedback = 0L;
    pChlExParms->ExitResponse = MQXCC_OK;
    WriteLogFile (605, 'D', "return(MQXCC_OK)\n");
    break;
  case MQXR_TERM:
    WriteLogFile (610, 'D', "ExitReason = MQXR_TERM\n");
    pChlExParms->ExitResponse = MQXCC_OK;
    WriteLogFile (605, 'D', "return(MQXCC_OK)\n");
    break;
  case MQXR_MSG:
    WriteLogFile (611, 'D', "ExitReason = MQXR_MSG\n");
    WriteLogFile (605, 'D', "return(MQXCC_OK)\n");
    break;
  case MQXR_XMIT:
    WriteLogFile (612, 'D', "ExitReason = MQXR_XMIT\n");
    WriteLogFile (605, 'D', "return(MQXCC_OK)\n");
    break;
  case MQXR_INIT_SEC:
    WriteLogFile (620, 'D', "ExitReason = MQXR_INIT_SEC\n");
    switch (pChDef->ChannelType) {
    case MQCHT_SENDER:
      WriteLogFile (621, 'D', "Channel type = MQCHT_SENDER\n");
      WriteLogFile (606, 'D', "return(MQXCC_SEND_SEC_MSG)\n");
      break;
    case MQCHT_SVRCONN:
      WriteLogFile (626, 'D', "Channel type = MQCHT_SVRCONN\n");
    case MQCHT_RECEIVER:
      WriteLogFile (622, 'D', "Channel type = MQCHT_RECEIVER\n");
      buf = AllocExitBuffer (MQS_HAM_BUFLEN);
      if (buf == NULL) {
        WriteLogFile (56, 'E', "Buffer allocation failed for ExitBuffer\n");
        pChlExParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
        pChlExParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
      }
      else {
        *exitBufAddr = buf;
        *exitBufLength = MQS_HAM_BUFLEN;
        status = InitHAM (&mqhamMsg);
        memcpy (&origMqhamMsg, &mqhamMsg, sizeof (MQ_HAM));
        status = BuildHAM (buf, &bufLen, &mqhamMsg);
        if (status != 0) {
          WriteLogFile (57, 'E',
                        "Error building the HAM. Error code: 0x%04lX\n",
                        status);
          pChlExParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
          pChlExParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;
        }
        else {
          *dataLength = bufLen;
          pChlExParms->ExitResponse = MQXCC_SEND_AND_REQUEST_SEC_MSG;
          pChlExParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;
          WriteLogFile (606, 'D', "return(MQXCC_SEND_AND_REQUEST_SEC_MSG)\n");
        }
      }
      break;
    case MQCHT_SERVER:
      WriteLogFile (623, 'D', "Channel type = MQCHT_SERVER\n");
      break;
    case MQCHT_REQUESTER:
      WriteLogFile (624, 'D', "Channel type = MQCHT_REQUESTER\n");
      break;
    case MQCHT_CLNTCONN:
      WriteLogFile (625, 'D', "Channel type = MQCHT_CLNTCONN\n");
      break;
    default:
      WriteLogFile (67, 'E', "Channel type = (unknown type)\n");
    }                                /* endswitch ChannelType in MQXR_INIT_SEC */
    break;
  case MQXR_SEC_MSG:
    WriteLogFile (630, 'D', "ExitReason = MQXR_SEC_MSG\n");
    switch (pChDef->ChannelType) {
    case MQCHT_SERVER:
      WriteLogFile (623, 'D', "Channel type = MQCHT_SERVER\n");
// break;
    case MQCHT_RECEIVER:
      WriteLogFile (622, 'D', "Channel type = MQCHT_RECEIVER\n");
      status = ReadAnswerHAM (&mqhamMsg, pagBuf, *dataLength);
      if (status != 0) {
        WriteLogFile (32, 'E',
                      "Sending channel exit failed to pass security check. Error code: 0x%04lX\n",
                      status);
        WriteLogFile (404, 'I',
                      "Channel will be closed by the receiver due to security violation\n");
        pChlExParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
        pChlExParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
        WriteLogFile (606, 'D', "return(MQXCC_CLOSE_CHANNEL)\n");
      }
      else {
        WriteLogFile (420, 'I',
                      "Sending channel exit passed security check successfully\n");
        pChlExParms->ExitResponse = MQXCC_OK;
        pChlExParms->ExitResponse2 = 0L;
        WriteLogFile (606, 'D', "return(MQXCC_OK)\n");
      };
      break;
    case MQCHT_CLNTCONN:
      WriteLogFile (625, 'D', "Channel type = MQCHT_CLNTCONN\n");
    case MQCHT_SENDER:
      WriteLogFile (621, 'D', "Channel type = MQCHT_SENDER\n");
      status = ReceiveHAM (&mqhamMsg, pagBuf, *dataLength);
      if (status != 0) {
        WriteLogFile (33, 'E',
                      "Receiving channel exit failed to pass security check. Error code: 0x%04lX\n",
                      status);
        WriteLogFile (404, 'I',
                      "Channel will be closed by the sender due to security violation\n");
        pChlExParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
        pChlExParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
        WriteLogFile (606, 'D', "return(MQXCC_CLOSE_CHANNEL)\n");
      }
      else {
        WriteLogFile (421, 'I',
                      "Receiving channel exit passed security check successfully\n");
        buf = AllocExitBuffer (MQS_HAM_BUFLEN);
        if (buf == NULL) {
          WriteLogFile (56, 'E', "Buffer allocation failed for ExitBuffer\n");
          pChlExParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
          pChlExParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
          WriteLogFile (606, 'D', "return(MQXCC_SUPPRESS_FUNCTION)\n");
        }
        else {
          *exitBufAddr = buf;
          *exitBufLength = MQS_HAM_BUFLEN;
          status = BuildHAM (buf, &bufLen, &mqhamMsg);
          if (status != 0) {
            WriteLogFile (57, 'E', "Error building the HAM\n");
            pChlExParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
            pChlExParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;
            WriteLogFile (606, 'D', "return(MQXCC_SUPPRESS_FUNCTION)\n");
          }
          else {
            *dataLength = bufLen;
            pChlExParms->ExitResponse = MQXCC_SEND_SEC_MSG;
            pChlExParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;
            WriteLogFile (606, 'D', "return(MQXCC_SEND_SEC_MSG)\n");
          }
        }
      }
      break;
    case MQCHT_REQUESTER:
      WriteLogFile (624, 'D', "Channel type = MQCHT_REQUESTER\n");
      break;
    case MQCHT_SVRCONN:
      WriteLogFile (626, 'D', "Channel type = MQCHT_SVRCONN\n");
      break;
    default:
      WriteLogFile (67, 'E', "Channel type = (unknown type)\n");
    }                                /* endswitch ChannelType */
    break;
  case MQXR_RETRY:
    WriteLogFile (631, 'D', "ExitReason = MQXR_RETRY\n");
    break;
  default:
    WriteLogFile (39, 'E',
                  "Unknown ExitReason in channel security exit call\n");
  }
  return;
}

/*
* This is the function for the send message exit.
*/
void MQENTRY SendMessageExit (PMQVOID channelExitParms,
                              PMQVOID channelDef,
                              PMQLONG dataLength,
                              PMQLONG agBufLength,
                              PMQVOID agBuf,
                              PMQLONG exitBufLength, PMQPTR exitBufAddr)
{
  PMQCXP pChlExParms = (PMQCXP) channelExitParms;
  PMQCD pChDef = (PMQCD) channelDef;
  unsigned char *buf;
  unsigned int bufLen;
  int status;
  switch (pChlExParms->ExitReason) {
  case MQXR_INIT:
    WriteLogFile (630, 'D', "ExitReason = MQXR_INIT\n");
    break;
  case MQXR_TERM:
    WriteLogFile (631, 'D', "ExitReason = MQXR_TERM\n");
    pChlExParms->ExitResponse = MQXCC_OK;
    pChlExParms->ExitResponse2 = MQXCC_OK;
    pChlExParms->Feedback = 0L;
    WriteLogFile (645, 'D', "return(MQXCC_OK)\n");
    break;
  case MQXR_MSG:
    WriteLogFile (640, 'D', "ExitReason = MQXR_MSG\n");
    break;
  case MQXR_XMIT:
    WriteLogFile (650, 'D', "ExitReason = MQXR_XMIT\n");
    switch (pChDef->ChannelType) {
    case MQCHT_CLNTCONN:
      WriteLogFile (625, 'D', "Channel type = MQCHT_CLNTCONN\n");
    case MQCHT_SENDER:
      WriteLogFile (621, 'D', "Channel type = MQCHT_SENDER\n");
      if (*dataLength < 300) {
        pChlExParms->ExitResponse = MQXCC_OK;
        pChlExParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
        WriteLogFile (666, 'D', "ExitResponse2 = MQXR2_USE_AGENT_BUFFER\n");
        WriteLogFile (606, 'D', "return(MQXCC_OK)\n");
        break;
      }
      buf = BuildBufferToSendMSG (&bufLen, &status, agBuf, *dataLength);
      if (status == 0) {
        if (buf) {
          *exitBufAddr = buf;
          *exitBufLength = bufLen;
          *dataLength = bufLen;
          WriteLogFile (632, 'D', "Sending buffer length: %d\n", *dataLength);
          pChlExParms->ExitResponse = MQXCC_OK;
          pChlExParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;
          WriteLogFile (666, 'D', "ExitResponse2 = MQXR2_USE_EXIT_BUFFER\n");
          WriteLogFile (606, 'D', "return(MQXCC_OK)\n");
        }
        else {
          pChlExParms->ExitResponse = MQXCC_OK;
          pChlExParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
          WriteLogFile (666, 'D', "ExitResponse2 = MQXR2_USE_AGENT_BUFFER\n");
          WriteLogFile (606, 'D', "return(MQXCC_OK)\n");
        }
      }
      break;
    case MQCHT_SERVER:
      WriteLogFile (623, 'D', "Channel type = MQCHT_SERVER\n");
      break;
    case MQCHT_RECEIVER:
      WriteLogFile (622, 'D', "Channel type = MQCHT_RECEIVER\n");
      break;
    case MQCHT_REQUESTER:
      WriteLogFile (624, 'D', "Channel type = MQCHT_REQUESTER\n");
      break;
    case MQCHT_SVRCONN:
      WriteLogFile (626, 'D', "Channel type = MQCHT_SVRCONN\n");
      break;
    default:
      WriteLogFile (67, 'E', "Channel type = (unknown type)\n");
    }                                /* endswitch ChannelType */
    break;
  case MQXR_SEC_MSG:
    WriteLogFile (651, 'D', "ExitReason = MQXR_SEC_MSG\n");
    break;
  case MQXR_INIT_SEC:
    WriteLogFile (652, 'D', "ExitReason = MQXR_INIT_SEC\n");
    break;
  case MQXR_RETRY:
    WriteLogFile (653, 'D', "ExitReason = MQXR_RETRY\n");
    break;
  default:
    WriteLogFile (40, 'E', "Unknown ExitReason in send message exit call\n");
  }
}

/*
* This is the function for the receive message exit
*/
void MQENTRY ReceiveMessageExit (PMQVOID channelExitParms,
                                 PMQVOID channelDef,
                                 PMQLONG dataLength,
                                 PMQLONG agBufLength,
                                 PMQVOID agBuf,
                                 PMQLONG exitBufLength, PMQPTR exitBufAddr)
{
  PMQCXP pChlExParms = (PMQCXP) channelExitParms;
  PMQCD pChDef = (PMQCD) channelDef;
  unsigned char *buf;
  unsigned int bufLen;
  int status;
  switch (pChlExParms->ExitReason) {
  case MQXR_INIT:
    WriteLogFile (660, 'D', "ExitReason = MQXR_INIT\n");
    break;
  case MQXR_TERM:
    WriteLogFile (661, 'D', "ExitReason = MQXR_TERM\n");
    pChlExParms->ExitResponse = MQXCC_OK;
    pChlExParms->ExitResponse2 = MQXCC_OK;
    pChlExParms->Feedback = 0L;
    WriteLogFile (650, 'D', "return(MQXCC_OK)\n");
    break;
  case MQXR_MSG:
    WriteLogFile (662, 'D', "ExitReason = MQXR_MSG\n");
    break;
  case MQXR_XMIT:
    WriteLogFile (663, 'D', "ExitReason = MQXR_XMIT\n");
    switch (pChDef->ChannelType) {
    case MQCHT_SENDER:
      WriteLogFile (621, 'D', "Channel type = MQCHT_SENDER\n");
      break;
    case MQCHT_SERVER:
      WriteLogFile (623, 'D', "Channel type = MQCHT_SERVER\n");
      break;
    case MQCHT_SVRCONN:
      WriteLogFile (626, 'D', "Channel type = MQCHT_SVRCONN\n");
// break;
    case MQCHT_RECEIVER:
      WriteLogFile (622, 'D', "Channel type = MQCHT_RECEIVER\n");
      WriteLogFile (632, 'D', "Receive buffer length: %d\n", *dataLength);
      if (*dataLength < 300) {
        pChlExParms->ExitResponse = MQXCC_OK;
        pChlExParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
        WriteLogFile (666, 'D', "ExitResponse2 = MQXR2_USE_AGENT_BUFFER\n");
        WriteLogFile (606, 'D', "return(MQXCC_OK)\n");
        break;
      }
      buf = ReceiveMSGBuffer (&bufLen, agBuf, &status, *dataLength);
      if (status == 0) {
        if (buf) {
          *exitBufAddr = buf;
          *exitBufLength = bufLen;
          *dataLength = bufLen;
          WriteLogFile (632, 'D', "Sending buffer length: %d\n", *dataLength);
          pChlExParms->ExitResponse = MQXCC_OK;
          pChlExParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;
          WriteLogFile (666, 'D', "ExitResponse2 = MQXR2_USE_EXIT_BUFFER\n");
          WriteLogFile (606, 'D', "return(MQXCC_OK)\n");
        }
        else {
          pChlExParms->ExitResponse = MQXCC_OK;
          pChlExParms->ExitResponse2 = MQXR2_USE_AGENT_BUFFER;
          WriteLogFile (666, 'D', "ExitResponse2 = MQXR2_USE_AGENT_BUFFER\n");
          WriteLogFile (606, 'D', "return(MQXCC_OK)\n");
        }
      }
      break;
    case MQCHT_REQUESTER:
      WriteLogFile (624, 'D', "Channel type = MQCHT_REQUESTER\n");
      break;
    case MQCHT_CLNTCONN:
      WriteLogFile (625, 'D', "Channel type = MQCHT_CLNTCONN\n");
      break;
    default:
      WriteLogFile (67, 'E', "Channel type = (unknown type )\n");
    }                                /* endswitch ChannelType */
    break;
  case MQXR_SEC_MSG:
    WriteLogFile (664, 'D', "ExitReason = MQXR_SEC_MSG\n");
    break;
  case MQXR_INIT_SEC:
    WriteLogFile (665, 'D', "ExitReason = MQXR_INIT_SEC\n");
    break;
  case MQXR_RETRY:
    WriteLogFile (666, 'D', "ExitReason = MQXR_RETRY\n");
    break;
  default:
    WriteLogFile (41, 'E',
                  "Unknown ExitReason in receive message exit call\n");
  }                                /* endswitch ExitReason for ExitType
                                   * MQXT_CHANNEL_MSG_EXIT */
}
void GetInfo (PMQCD pChDef)
{
  MakeCString (channeldata.channel, pChDef->ChannelName,
               MQ_CHANNEL_NAME_LENGTH);
/* In case of a client connection channel we have no queue manager name
* We take the name provided in the MsgUserData field */
  if (pChDef->QMgrName[0] != 0x00 && pChDef->QMgrName[0] != ' ')
    MakeCString (qmgrdata.qmgr, pChDef->QMgrName, MQ_Q_MGR_NAME_LENGTH);
  else
    MakeCString (qmgrdata.qmgr, pChDef->SendUserData, 32);
/* Get hostname*/
  if (GetHostName (hostName, &hostNameLen) != 0)
    WriteLogFile (56, 'E', "Function GetHostName() was unsuccessful\n");
  if (memcmp (pChDef->SecurityUserData, "DEBUG", 5) == 0)
    doDebugging = 1;
  else
    doDebugging = 0;
  if (GetEnvironmentVariable (LOGFILE_PATH_VAR, logFilePath, 256) == 0) {
    WriteLogFile (43, 'E', "Environment variable %s not defined\n", LOGFILE_PATH_VAR);
  }
  if (GetEnvironmentVariable (KEYFILE_PATH_VAR, keyFilePath, 256) == 0) {
    WriteLogFile (44, 'E', "Environment variable %s not defined\n", KEYFILE_PATH_VAR);
  }
}
void WriteLogFile (int msgid, unsigned char msgtype, unsigned char *msgtxt, ...)
{
  time_t ltime;
  struct tm *newtime;
  static int firstTime = 1;
  va_list args;
  time (&ltime);

  newtime = localtime (&ltime);
  if (firstTime) {
    firstTime = 0;
    WriteLogFile (400, 'I', "*********** ChannelExit started ***********\n");
    printf ("MQS9400I %s WT05219A MQCHEXIT ChannelExit started\n", hostName);
  }
  if (doDebugging == 0 && msgtype == 'D')
    return;
  if (logFilePtr != stdout)
    if ((logFilePtr = fopen (logFileName, "a+")) == NULL) {
      logFilePtr = stdout;
      WriteLogFile (40, 'E', "Error opening logfile %s.\n", logFileName);
      WriteLogFile (441, 'I', "Using stdout for log messages\n");
      return;
    }
  fprintf (logFilePtr, "%02d%03d %02d:%02d:%02d ", newtime->tm_year,
           newtime->tm_yday, newtime->tm_hour, newtime->tm_min,
           newtime->tm_sec);
  fprintf (logFilePtr, "MQS9%03d%c %s %s MQCHEXIT ", msgid, msgtype, hostName,
           qmgrdata.qmgr);
  va_start (args, msgtxt);
  vfprintf (logFilePtr, msgtxt, args);
  va_end (args);
  if (logFilePtr != stdout)
    fclose (logFilePtr);
  else
    fflush (stdout);
#ifdef _MVS
  wto (errstr);
#endif
}

/* Builds the HAM (Handshaking Authentification message) to transmit to
* the sending queue manager' s channelexit.
* Return the len of the HAM.
*/
int BuildHAM (char *buf, unsigned int *bufLen, MQ_HAM * mqham)
{
  int status;
  WriteLogFile (700, 'D',
                "Encrypting the Handshaking Authentication Message\n");
  status = EncryptHAM (mqham, &partnerpubk, &qmgrdata.privk);
  if (status != 0)
    return status;
  WriteLogFile (705, 'D', "Building the buffer to send the HAM\n");
  status = BuildBufferToSendHAM (buf, bufLen, mqham);
  return status;
}

/* Sending queue manager' s channel security exit has to read HAM and to
answer properly.
* If ok it returns 0, otherwise (e.g. if HAM is not in proper format)
it returns -1.
*/
int ReceiveHAM (MQ_HAM * mqham, char *buf, unsigned int buflen)
{
  int status;
  if (buf == NULL)
    return -1;
  WriteLogFile (720, 'D',
                "Receiving the Handshaking Authentication Message\n");
  status = ReadHAM (mqham, buf, buflen);
  if (status != 0)
    return status;
  WriteLogFile (725, 'D', "Verifying the HAM\n");
  status = VerifyHAM (mqham, &partnerpubk, &qmgrdata.privk);
  return status;
}
int ReadAnswerHAM (MQ_HAM * mqham, char *buf, unsigned int buflen)
{
  int status;
  if (buf == NULL)
    return -1;
  WriteLogFile (730, 'D',
                "Receiving the answer to the Handshaking Authentication Message\n");
  status = ReadHAM (mqham, buf, buflen);
  if (status != 0)
    return status;
  WriteLogFile (735, 'D', "Verifying the answer to the HAM\n");
  status =
    VerifyAnswerHAM (mqham, &origMqhamMsg, &partnerpubk, &qmgrdata.privk);
  if (status != 0)
    return status;
  if (memcmp (mqham->hambuf, origMqhamMsg.hambuf, mqham->hambuflen) == 0)
    return 0;
  else
    return -1;
  return 0;
}
unsigned char *AllocExitBuffer (int len)
{
  unsigned char *buf;
  buf = (unsigned char *) malloc (len);
  if (buf == NULL)
    return NULL;
  return buf;
}

void FreeExitBuffer (PMQPTR exitBufAddr, PMQLONG exitBufLength)
{
  if (exitBufAddr != NULL && *exitBufAddr != NULL) {
    free (*exitBufAddr);
    *exitBufAddr = NULL;
    *exitBufLength = 0L;
  }
}
int ReadKeysFromFiles (void)
{
  int status;
  if ((status = ReadPrivateKey ()) != 0)
    return status;
  WriteLogFile (700, 'D', "Reading private keys was successful\n");
  if ((status = ReadPublicKey ()) != 0)
    return status;
  WriteLogFile (701, 'D', "Reading public keys was successful\n");
  return 0;
}
int ReadPrivateKey (void)
{
  FILE *fptr;
  unsigned char qmname[MQ_Q_MGR_NAME_LENGTH];
/* Build name of file which does contain the private key */
  strcpy (privKeyFileName, keyFilePath);
  strcat (privKeyFileName, qmgrdata.qmgr);
  strcat (privKeyFileName, ".PRV");
  WriteLogFile (690, 'D', "Opening private key file\n");
  fptr = fopen (privKeyFileName, "rb");
  if (fptr == NULL)
    return 2;
/* Read queue manager' s name from file */
  fread (tmpBuf, QMGR_NAME_EYE_CATCHER_SIZE, 1, fptr);
  if (memcmp (tmpBuf, QMGR_NAME_EYE_CATCHER, QMGR_NAME_EYE_CATCHER_SIZE) != 0)
    return 4;
  fread (qmname, MQ_Q_MGR_NAME_LENGTH, 1, fptr);
/* Check if name is ok */
  if (memcmp (qmgrdata.qmgr, qmname, strlen (qmgrdata.qmgr)) != 0)
    return 6;
/* Read the CSSD's public key from file */
  fread (tmpBuf, CSSD_PUBK_EYE_CATCHER_SIZE, 1, fptr);
  if (memcmp (tmpBuf, CSSD_PUBK_EYE_CATCHER, CSSD_PUBK_EYE_CATCHER_SIZE) != 0)
    return 8;
  fread (&cssdPubk, sizeof (R_RSA_PUBLIC_KEY), 1, fptr);
/* Read the queue manager' s public key from file */
  fread (tmpBuf, QMGR_PUBK_EYE_CATCHER_SIZE, 1, fptr);
  if (memcmp (tmpBuf, QMGR_PUBK_EYE_CATCHER, QMGR_PUBK_EYE_CATCHER_SIZE) != 0)
    return 10;
  fread (&qmgrdata.pubk, sizeof (R_RSA_PUBLIC_KEY), 1, fptr);
/* Read the queue manager' s private key from file */
  fread (tmpBuf, QMGR_PRIVK_EYE_CATCHER_SIZE, 1, fptr);
  if (memcmp (tmpBuf, QMGR_PRIVK_EYE_CATCHER, QMGR_PRIVK_EYE_CATCHER_SIZE) !=
      0) return 12;
  fread (&qmgrdata.privk, sizeof (R_RSA_PRIVATE_KEY), 1, fptr);
  fclose (fptr);
  return 0;
}
int ReadPublicKey (void)
{
  FILE *fptr;
  unsigned char qmname[MQ_Q_MGR_NAME_LENGTH];
  MQCHAR channelname[MQ_CHANNEL_NAME_LENGTH];
  int found = 0;
  unsigned char signature[MAX_SIGNATURE_LEN];
  unsigned int signatureLen;
  R_SIGNATURE_CTX context;
  int status;
  int offset;
/* Build name of file which does contain the private key */
  strcpy (pubKeyFileName, keyFilePath);
  strcat (pubKeyFileName, qmgrdata.qmgr);
  strcat (pubKeyFileName, ".PUB");
  WriteLogFile (691, 'D', "Opening public key file\n");
  fptr = fopen (pubKeyFileName, "rb");
  if (fptr == NULL)
    return 32;
  while (fread (tmpBuf, CHANNEL_NAME_EYE_CATCHER_SIZE, 1, fptr)) {
    if (memcmp (tmpBuf, CHANNEL_NAME_EYE_CATCHER,
                CHANNEL_NAME_EYE_CATCHER_SIZE) != 0)
      return 34;
    fread (channelname, MQ_CHANNEL_NAME_LENGTH, 1, fptr);
    if (memcmp (channelname, channeldata.channel,
                strlen (channeldata.channel)) == 0) {
/* We found the channel in the file
* Reading partner info: name of qmgr and public key now. */
      fread (tmpBuf, QMGR_NAME_EYE_CATCHER_SIZE, 1, fptr);
      if (memcmp (tmpBuf, QMGR_NAME_EYE_CATCHER,
                  QMGR_NAME_EYE_CATCHER_SIZE) != 0)
        return 36;
      fread (partnername, MQ_Q_MGR_NAME_LENGTH, 1, fptr);
      fread (tmpBuf, QMGR_PUBK_EYE_CATCHER_SIZE, 1, fptr);
      if (memcmp (tmpBuf, QMGR_PUBK_EYE_CATCHER,
                  QMGR_PUBK_EYE_CATCHER_SIZE) != 0)
        return 38;
      fread (&partnerpubk, sizeof (R_RSA_PUBLIC_KEY), 1, fptr);
/* Reading the digital signature to verify the info just gotten */
      fread (tmpBuf, DIGSIG_EYE_CATCHER_SIZE, 1, fptr);
      if (memcmp (tmpBuf, DIGSIG_EYE_CATCHER, DIGSIG_EYE_CATCHER_SIZE) != 0)
        return 40;
      fread (&signatureLen, sizeof (unsigned int), 1, fptr);
      fread (signature, MAX_SIGNATURE_LEN, 1, fptr);
/* We will verify the data now */
      R_VerifyInit (&context, DA_MD5);
      R_VerifyUpdate (&context, channelname, MQ_CHANNEL_NAME_LENGTH);
      R_VerifyUpdate (&context, partnername, MQ_Q_MGR_NAME_LENGTH);
      R_VerifyUpdate (&context, (unsigned char *) &partnerpubk,
                      sizeof (R_RSA_PUBLIC_KEY));
      status = R_VerifyFinal (&context, signature, signatureLen, &cssdPubk);
      if (status != 0)
        return status;
      found = 1;                /* Indicate success */
      break;
    }
    else {
/* point to next record */
      offset = QMGR_NAME_EYE_CATCHER_SIZE + MQ_Q_MGR_NAME_LENGTH +
        QMGR_PUBK_EYE_CATCHER_SIZE + sizeof (R_RSA_PUBLIC_KEY) +
        DIGSIG_EYE_CATCHER_SIZE + sizeof (unsigned int) + MAX_SIGNATURE_LEN;
      fseek (fptr, offset, SEEK_CUR);
    }
  }
  if (found == 0)
    return 50;
  fclose (fptr);
  return 0;
}
void MQStart (void)
{
}

/* Copy from a blank-filled array to a zero-terminated string */
void MakeCString (char *zt, char *bf, int len)
{
  int i;
  for (i = 0; bf[i] != ' ' && i < len; i++)
    zt[i] = bf[i];
  zt[i] = 0;
}
void BuildLogFileName (void)
{
/* Build logfile name */
  strcpy (logFileName, logFilePath);
  strcat (logFileName, qmgrdata.qmgr);
  strcat (logFileName, ".MQCHEXIT.LOG");
  return;
}

int InitHAM (MQ_HAM * mqham)
{
  FillRandomBuffer16 (mqham->hambuf);
  mqham->hambuflen = 16;
  return 0;
}

/* Caller has to provide a 32 bytes outbuf for the HAM to be put to.
*/
int EncryptHAM (MQ_HAM * mqham,
                R_RSA_PUBLIC_KEY * pubk, R_RSA_PRIVATE_KEY * privk)
{
  int status;
  status = EncryptBuffer (mqham->hamEncryptedBuf,
                          &(mqham->hamEncryptedBufLen),
                          mqham->hambuf,
                          mqham->hambuflen,
                          MQ_DO_NO_COMPRESS, pubk, privk, &(mqham->mqdigsig));
  return status;
}
int BuildBufferToSendHAM (unsigned char *outbuf,
                          unsigned int *outbuflen, MQ_HAM * mqham)
{
  memcpy (outbuf, HAM_EYE_CATCHER, HAM_EYE_CATCHER_SIZE);
  *outbuflen = HAM_EYE_CATCHER_SIZE;
  memcpy (outbuf + *outbuflen, &(mqham->mqdigsig), sizeof
          (MQ_DIGITAL_SIGNATURE));
  *outbuflen += sizeof (MQ_DIGITAL_SIGNATURE);
  memcpy (outbuf + *outbuflen, &(mqham->hamEncryptedBufLen),
          sizeof (unsigned int));
  *outbuflen += sizeof (unsigned int);
  memcpy (outbuf + *outbuflen, mqham->hamEncryptedBuf,
          mqham->hamEncryptedBufLen);
  *outbuflen += mqham->hamEncryptedBufLen;
  WriteLogFile (607, 'D', "Sending %d bytes. Encypted buffer length %d\n",
                *outbuflen, mqham->hamEncryptedBufLen);
  return 0;
}

/* Returns an encrypted digital signature and the encrypted buffer in outbuf.
* Caller must provide a buffer large enough to hold the data.
* The function itself returns the number of bytes written to outbuf in oubbuflen.
* ATTENTION: Currently compression is not implemented.
*/
int EncryptBuffer (unsigned char *outbuf,
                   unsigned int *outbuflen,
                   unsigned char *inbuf,
                   unsigned int inbuflen,
                   unsigned int compressflag,
                   R_RSA_PUBLIC_KEY * recpubk,
                   R_RSA_PRIVATE_KEY * sdrprivk,
                   MQ_DIGITAL_SIGNATURE * mqdigsig)
{
  R_RANDOM_STRUCT random;
  R_ENVELOPE_CTX context;
  int status;
  unsigned char encryptedKey[MAX_ENCRYPTED_KEY_LEN];
  unsigned char *encryptedKeys[1];
  unsigned int encryptedKeyLen;
  unsigned char iv[8];
  unsigned char signature[MAX_SIGNATURE_LEN];
  unsigned int signatureLen;
  unsigned int outlen;
  unsigned char *partOut;
  unsigned int partOutLen;
  unsigned char *partIn;
  unsigned int partInLen;
  status = 0;
  R_RandomCreate (&random);
  encryptedKeys[0] = encryptedKey;
  R_memset ((unsigned char *) &context, 0, sizeof (R_ENVELOPE_CTX));
/* Digital signature of block */
  status = R_SignBlock (signature, &signatureLen, inbuf, inbuflen, DA_MD5,
                        sdrprivk);
  mqdigsig->signatureLen = signatureLen;
  memcpy (mqdigsig->signature, signature, signatureLen);
  status = R_SealInit (&context, encryptedKeys, &encryptedKeyLen, iv, 1,        /* number of public keys */
                       &recpubk, EA_DES_EDE3_CBC, &random);
  if (status != 0)
    return status;
  partIn = inbuf;
  partInLen = 2048;
  partOut = outbuf;
  partOutLen = 0;
  outlen = 0;
  while (inbuflen > 0) {
    if (partInLen > inbuflen)
      partInLen = inbuflen;
    status = R_SealUpdate (&context, partOut, &partOutLen, partIn, partInLen);
    if (status != 0)
      return status;
    outlen += partOutLen;
    partOut += partOutLen;
    inbuflen -= partInLen;
    partIn += partInLen;
  }
  status = R_SealFinal (&context, partOut, &partOutLen);
  if (status != 0)
    return status;
  outlen += partOutLen;
  partOut += partOutLen;
  *outbuflen = outlen;
  R_RandomFinal (&random);
  mqdigsig->encryptedKeyLen = encryptedKeyLen;
  memcpy (mqdigsig->encryptedKey, encryptedKey, encryptedKeyLen);
  memcpy (mqdigsig->iv, iv, sizeof (iv));
/* Clear sensitive information. */
  R_memset ((POINTER) & context, 0, sizeof (context));
  R_memset ((POINTER) inbuf, 0, inbuflen);
  return 0;
}

/* Checks if the encrypted buffer inbuf goes along with the digital signature
* mqdigsig and decrypts the content of inbuf to outbuf. The number of bytes
* written to outbuf are returned in outbuflen.
* Caller must provide a buffer large enough to hold thee output data.
*/
int DecryptAndVerifyBuffer (unsigned char *outbuf,
                            unsigned int *outbuflen,
                            unsigned char *inbuf,
                            unsigned int inbuflen,
                            R_RSA_PUBLIC_KEY * sdrpubk,
                            R_RSA_PRIVATE_KEY * recprivk,
                            MQ_DIGITAL_SIGNATURE * mqdigsig)
{
  R_RANDOM_STRUCT random;
  R_ENVELOPE_CTX context;
  int status;
  unsigned char encryptedKey[MAX_ENCRYPTED_KEY_LEN];
  unsigned char *encryptedKeys[1];
  unsigned int encryptedKeyLen;
  unsigned char iv[8];
  unsigned int inlen;
  unsigned int outlen;
  ULONG magic;
  unsigned char *partOut;
  unsigned int partOutLen;
  unsigned char *partIn;
  unsigned int partInLen;
  USHORT slen;
  status = 0;
  status = R_OpenInit (&context,
                       EA_DES_EDE3_CBC,
                       mqdigsig->encryptedKey,
                       mqdigsig->encryptedKeyLen, mqdigsig->iv, recprivk);
  if (status != 0)
    return status;
  partIn = inbuf;
  partInLen = 1024;
  partOut = outbuf;
  partOutLen = 0;
  outlen = 0;
  while (inbuflen > 0) {
    if (partInLen > inbuflen)
      partInLen = inbuflen;
    status = R_OpenUpdate (&context, partOut, &partOutLen, partIn, partInLen);
    if (status != 0)
      return status;
    outlen += partOutLen;
    partOut += partOutLen;
    inbuflen -= partInLen;
    partIn += partInLen;
  };
  status = R_OpenFinal (&context, partOut, &partOutLen);
  if (status != 0)
    return status;
  outlen += partOutLen;
  partOut += partOutLen;
  *outbuflen = outlen;
/* Verify digital signature of block */
  status = R_VerifyBlockSignature (outbuf, *outbuflen,
                                   mqdigsig->signature,
                                   mqdigsig->signatureLen, DA_MD5, sdrpubk);
  if (status != 0)
    return status;
/* Clear sensitive information. */
  R_memset ((POINTER) & context, 0, sizeof (context));
  return 0;
}

/* Fills a buffer of len bytes with random values. Buffer must be
provided by caller.
* Returns 0 if successful, -1 otherwise.
*/
int FillRandomBuffer (unsigned char *outbuf, int len)
{
  R_RANDOM_STRUCT randomstruct;
  if (outbuf == 0)
    return -1;
  R_RandomCreate (&randomstruct);
  R_GenerateBytes (outbuf, len, &randomstruct);
  R_RandomFinal (&randomstruct);
  return 0;
}

/* Fills a 16 bytes length buffer with random values. Buffer must be
provided by caller.
* Returns 0 if successful, -1 otherwise.
*/
int FillRandomBuffer16 (unsigned char *outbuf)
{
  return FillRandomBuffer (outbuf, 16);
}
int ReadHAM (MQ_HAM * mqham, unsigned char *rcvrbuf, unsigned int rcvrbuflen)
{
  int offset;
  memcpy (tmpBuf, rcvrbuf, HAM_EYE_CATCHER_SIZE);
  offset = HAM_EYE_CATCHER_SIZE;
  if (memcmp (tmpBuf, HAM_EYE_CATCHER, HAM_EYE_CATCHER_SIZE) != 0)
    return 100;
  memcpy (&(mqham->mqdigsig), rcvrbuf + offset,
          sizeof (MQ_DIGITAL_SIGNATURE));
  offset += sizeof (MQ_DIGITAL_SIGNATURE);
  memcpy (&(mqham->hamEncryptedBufLen), rcvrbuf + offset,
          sizeof (unsigned int));
  offset += sizeof (unsigned int);
  memcpy (mqham->hamEncryptedBuf, rcvrbuf + offset,
          mqham->hamEncryptedBufLen);
  WriteLogFile (700, 'D', "%d bytes read. Encypted buffer length is %d\n",
                offset + mqham->hamEncryptedBufLen,
                mqham->hamEncryptedBufLen);
  return 0;
}

int VerifyHAM (MQ_HAM * mqham,
               R_RSA_PUBLIC_KEY * pubk, R_RSA_PRIVATE_KEY * privk)
{
  int status;
  status = DecryptAndVerifyBuffer (mqham->hambuf,
                                   &(mqham->hambuflen),
                                   mqham->hamEncryptedBuf,
                                   mqham->hamEncryptedBufLen,
                                   pubk, privk, &(mqham->mqdigsig));
  return status;
}

int VerifyAnswerHAM (MQ_HAM * mqham, MQ_HAM * origmqham,
                     R_RSA_PUBLIC_KEY * sdrpubk,
                     R_RSA_PRIVATE_KEY * rcvrprivk)
{
  int status;
  status = VerifyHAM (mqham, sdrpubk, rcvrprivk);
  if (status != 0)
    return status;
/* Check that the message just decrypted is equal to what we did send in
* the beginning of the handshaking!
*/
  if (mqham->hambuflen != origmqham->hambuflen)
    return MQRE_HAM_MSGLEN;
  if (memcmp (mqham->hambuf, origmqham->hambuf, origmqham->hambuflen) != 0)
    return MQRE_HAM_MSG;
  return status;
}

#define FILLER ' '
#define HEX_PER_LINE 32
void PrintBuff (unsigned char *out, unsigned int len)
{
  int i,
    j;
  unsigned uNum = 0;
  if (doDebugging == 0)
    return;
  if (logFilePtr != stdout)
    if ((logFilePtr = fopen (logFileName, "a+")) == NULL) {
      logFilePtr = stdout;
      WriteLogFile (40, 'E', "Error opening logfile %s.\n", logFileName);
      WriteLogFile (41, 'E', "Using stdout for log messages\n");
      return;
    }
  for (i = 0; i < len; i += HEX_PER_LINE) {
    fprintf (logFilePtr, "%5d %4x:", uNum, uNum);
    for (j = 0; j < HEX_PER_LINE; j++) {
      if (j % 4 == 0)
        fprintf (logFilePtr, " ");
      if (j + i < len)
        fprintf (logFilePtr, "%02x", (int) out[j + i]);
      else
        fprintf (logFilePtr, " ");
    }
    fprintf (logFilePtr, " , ");
    for (j = 0; j < HEX_PER_LINE; j++) {
      if (j + i < len) {
        if (isprint ((int) out[j + i]) == 0)
          fprintf (logFilePtr, "%c", FILLER);
        else
          fprintf (logFilePtr, "%c", (int) out[j + i]);
      }
      else
        fprintf (logFilePtr, " ");
    }
    fprintf (logFilePtr, "*\n");
    uNum += HEX_PER_LINE;
  }
  if (logFilePtr != stdout)
    fclose (logFilePtr);
  else
    fflush (stdout);
}
unsigned char *BuildBufferToSendMSG (unsigned int *outbuflen,
                                     int *status,
                                     unsigned char *inbuf,
                                     unsigned int inbuflen)
{
  unsigned char *buffer;
  unsigned char *optr;
  unsigned int offsetOut;
  unsigned int offsetIn;
  unsigned int encryptBufLen;
  MQ_DIGITAL_SIGNATURE mqdigsig;
  int mallocSz;
  *status = 0;
  mallocSz = 8 + MSG_EYE_CATCHER_SIZE + sizeof (MQ_DIGITAL_SIGNATURE) +
    sizeof (unsigned int) + inbuflen + 1 + 8;
  if ((buffer = (unsigned char *) malloc (mallocSz)) == 0)
    return 0;
  WriteLogFile (923, 'D', "Malloc size %d\n", mallocSz);
  memcpy (buffer, inbuf, 8);
  offsetOut = 8;
  offsetIn = 8;
  inbuflen -= offsetIn;
  buffer[8] = 0;                /* This is to indicate that data is encrypted and
                                   should be handled by the receiver' s side */
  offsetOut += 1;
  memcpy (buffer + offsetOut, MSG_EYE_CATCHER, MSG_EYE_CATCHER_SIZE);
  offsetOut += MSG_EYE_CATCHER_SIZE;
  optr = buffer + offsetOut + sizeof (MQ_DIGITAL_SIGNATURE) +
    sizeof (unsigned int);
  *status = EncryptBuffer (optr,
                           &encryptBufLen,
                           inbuf + offsetIn,
                           inbuflen - 8,
                           MQ_DO_NO_COMPRESS,
                           &partnerpubk, &qmgrdata.privk, &mqdigsig);
  memcpy (buffer + 9 + MSG_EYE_CATCHER_SIZE, &mqdigsig,
          sizeof (MQ_DIGITAL_SIGNATURE));
  memcpy (buffer + 9 + MSG_EYE_CATCHER_SIZE +
          sizeof (MQ_DIGITAL_SIGNATURE), &encryptBufLen,
          sizeof (unsigned int));
  *outbuflen =
    8 + 1 + MSG_EYE_CATCHER_SIZE + sizeof (MQ_DIGITAL_SIGNATURE) +
    sizeof (unsigned int) + encryptBufLen;
// PrintBuff(buffer,*outbuflen);
  return buffer;
}
unsigned char *ReceiveMSGBuffer (unsigned int *outbuflen,
                                 unsigned char *inbuf,
                                 int *status, unsigned int inbuflen)
{
  unsigned char *buffer;
  unsigned char *optr;
  unsigned int offsetOut;
  unsigned int offsetIn;
  unsigned int encryptBufLen;
  MQ_DIGITAL_SIGNATURE mqdigsig;
  unsigned int inlen;
  *status = 0;
  if (inbuf[8] != 0)
    return 0;
// PrintBuff(inbuf,inbuflen);
  if (memcmp (inbuf + 9, MSG_EYE_CATCHER, MSG_EYE_CATCHER_SIZE) != 0)
    return 0;
  if ((buffer = (unsigned char *) malloc (inbuflen + 8)) == 0)
    return 0;
  memcpy (buffer, inbuf, 8);
  memcpy (&mqdigsig, inbuf + 9 + MSG_EYE_CATCHER_SIZE,
          sizeof (MQ_DIGITAL_SIGNATURE));
// PrintBuff((unsigned char*) &mqdigsig, sizeof(MQ_DIGITAL_SIGNATURE) );
  memcpy (&inlen, inbuf + 8 + 1 + MSG_EYE_CATCHER_SIZE +
          sizeof (MQ_DIGITAL_SIGNATURE), sizeof (unsigned int));
  WriteLogFile (890, 'D', "Dumping 'unsigned int inlen' value\n");
  PrintBuff ((unsigned char *) &inlen, sizeof (unsigned int));
  *status = DecryptAndVerifyBuffer (buffer + 8,
                                    &encryptBufLen,
                                    inbuf + 1 + 8 + MSG_EYE_CATCHER_SIZE +
                                    sizeof (MQ_DIGITAL_SIGNATURE) +
                                    sizeof (unsigned int), inlen,
                                    &partnerpubk, &qmgrdata.privk, &mqdigsig);
  *outbuflen = encryptBufLen + 8;
// PrintBuff(buffer, *outbuflen );
  return buffer;
}

/* Returns 0 if successful, otherwise != 0.
* If successful the hostname is returned in buf and its length in bufLen.
* ATTENTION: The application has to be linked with WSOCK32.LIB.
*/
int GetHostName (unsigned char *buf, unsigned int *bufLen)
{
  int status;
  WSADATA wsaData;
  WORD wVersionRequested;
  wVersionRequested = MAKEWORD (1, 1);
  status = WSAStartup (wVersionRequested, &wsaData);
  if (status != 0)
    return status;
  status = gethostname (buf, 255);
  if (status != 0)
    return status;
  *bufLen = strlen (buf) + 1;        /* we want to count the ending zero */
  return 0;
}
